from . import left, left_short, right, right_short, sencoder
import pandas as pd


def test_encoding():
    encoded = sencoder(left, right.iloc[:-3, :])
    assert sencoder.similarity_map.no_associations() == len(
        encoded
    ), f"Expected {sencoder.similarity_map.no_associations()}, got {len(encoded)}"
    for i, v in enumerate(encoded):
        assert v.shape[-1] == sencoder.similarity_map.association_sizes()[i], (
            f"Expected {sencoder.similarity_map.association_sizes()[i]}, "
            f"got {v.shape[-1]}"
        )


def test_encoding_report():
    report = sencoder.report_encoding(left_short, right_short)
    assert isinstance(report, list), "Failed to report encoding."
    expected_shape = (len(sencoder.similarity_map), len(left_short))
    for item in report:
        assert isinstance(item, pd.DataFrame), "Failed to report encoding."
        assert item.shape == expected_shape, (
            "Unexpected report shape. " f"Expected: {expected_shape}, got: {item.shape}"
        )



from . import (
    dl_model,
    left_short,
    matches_short,
    ns_model,
    right_short,
    smap,
)
from neer_match.matching_model import DLMatchingModel, NSMatchingModel
import pytest
import numpy as np
import pandas as pd


def test_initialization():
    """Test initialization of the model."""
    with pytest.raises(ValueError):
        DLMatchingModel("no-similarity-map")
    with pytest.raises(ValueError):
        NSMatchingModel(smap, initial_record_width_scale=-1)
    with pytest.raises(ValueError):
        DLMatchingModel(smap, record_depth=-1)


def test_compilation():
    """Test initialization of the model."""
    dl_model.compile(loss="binary_crossentropy")
    assert dl_model is not None, "Failed to compile the model."
    ns_model.compile()
    assert ns_model is not None, "Failed to compile the model."


def test_fit():
    """Test fitting the model."""
    dl_model.fit(left_short, right_short, matches_short, epochs=1, batch_size=10)
    assert dl_model is not None, "Failed to fit the model."
    ns_model.fit(left_short, right_short, matches_short, epochs=1, batch_size=10)
    assert ns_model is not None, "Failed to fit the model."
    with pytest.raises(ValueError):
        ns_model.fit(None, right_short, matches_short, epochs=1, batch_size=0)
    with pytest.raises(ValueError):
        ns_model.fit(left_short, None, matches_short, epochs=1, batch_size=0)
    with pytest.raises(ValueError):
        ns_model.fit(left_short, right_short, None, epochs=1, batch_size=0)
    with pytest.raises(ValueError):
        ns_model.fit(left_short, right_short, matches_short, epochs=1.0, batch_size=0)
    with pytest.raises(ValueError):
        ns_model.fit(left_short, right_short, matches_short, epochs=1, batch_size=-1)


def test_evaluate():
    """Test evaluating the model."""
    results = ns_model.evaluate(left_short, right_short, matches_short)
    assert isinstance(results, dict), "Failed to evaluate the model."
    results = dl_model.evaluate(left_short, right_short, matches_short)
    assert isinstance(results, float), "Failed to evaluate the model."


def test_predict():
    """Test predicting with the model."""
    dl_preds = dl_model.predict(left_short, right_short)
    assert isinstance(
        dl_preds, np.ndarray
    ), "Failed to predict with the deep-learning model."
    expected_shape = (len(left_short) * len(right_short), 1)
    assert dl_preds.shape == expected_shape, (
        "Unexpected prediction shape. "
        f"Expected: {expected_shape}, got: {dl_preds.shape}"
    )
    ns_preds = ns_model.predict(left_short, right_short)
    assert isinstance(
        ns_preds, np.ndarray
    ), "Failed to predict with the neural-symbolic model."
    assert ns_preds.shape == expected_shape, (
        "Unexpected prediction shape. "
        f"Expected: {expected_shape}, got: {ns_preds.shape}"
    )


def test_suggest():
    """Test suggesting with the model."""
    dl_suggestions = dl_model.suggest(left_short, right_short, count=3)
    assert isinstance(
        dl_suggestions, pd.DataFrame
    ), "Failed to suggest with the deep-learning model."
    expected_shape = (len(left_short) * 3, 3)
    assert dl_suggestions.shape == expected_shape, (
        "Unexpected suggestion shape. "
        f"Expected: {expected_shape}, got: {dl_suggestions.shape}"
    )
    ns_suggestions = ns_model.suggest(left_short, right_short, count=3)
    assert isinstance(
        ns_suggestions, pd.DataFrame
    ), "Failed to suggest with the neural-symbolic model."
    assert ns_suggestions.shape == expected_shape, (
        "Unexpected suggestion shape. "
        f"Expected: {expected_shape}, got: {ns_suggestions.shape}"
    )



"""
Field pair network module.

This module contains functionality for instantiating, training, and using a field pair
matching networks.
"""

import tensorflow as tf


class FieldPairNetwork(tf.keras.Model):
    """Field pair network class.

    The class creates networks for matching pairs of fields from two datasets.

    Attributes:
        size (int): The size of the input feature vectors.
        initial_width_scale (int): The initial width scale of the hidden layers.
        depth (int): The depth of the network.
    """

    def __init__(
        self, size: int, initial_width_scale: int = 10, depth: int = 2, **kwargs
    ) -> None:
        """Initialize a field pair network object.

        The network depth is determined from the depth parameter. The width of each
        hidden layer is determined by the initial width scale and the number of input
        features. The first hidden layer has a width of size * initial_width_scale, the
        second has a width of size * initial_width_scale / 2, and so on. The output
        layer has a sigmoid activation function.

        Args:
            size: The size of the input feature vectors.
            initial_width_scale: The initial width scale of the hidden layers.
            depth: The depth of the network.
            **kwargs: Additional keyword arguments passed to parent class
                      (:func:`tensorflow.keras.Model`).
        """
        if not isinstance(size, int) or size < 1:
            raise ValueError("Size must be a positive integer.")
        if not isinstance(initial_width_scale, int) or initial_width_scale < 1:
            raise ValueError("Initial width scale must be a positive integer.")
        if not isinstance(depth, int) or depth < 1:
            raise ValueError("Depth must be a positive integer.")

        self.size = size
        self.initial_width_scale = initial_width_scale
        self.depth = depth
        super().__init__(**kwargs)

        self.field_layers = []
        for i in range(depth):
            self.field_layers += [
                tf.keras.layers.Dense(
                    max(int(initial_width_scale * size / (i + 1)), 2),
                    activation=tf.keras.activations.relu,
                    name=f"{self.name.replace('~', '_')}_hidden_{i}",
                )
            ]
        self.field_layers += [
            tf.keras.layers.Dense(
                1,
                tf.keras.activations.sigmoid,
                name=f"{self.name}_classifier",
            )
        ]

    def get_config(self) -> dict:
        """Return the network configuration."""
        config = super().get_config().copy()
        config.update(
            {
                "size": self.size,
                "initial_width_scale": self.initial_width_scale,
                "depth": self.depth,
            }
        )
        return config

    def build(self, input_shape : tf.TensorShape) -> None:
        """Build the network."""
        for layer in self.field_layers:
            layer.build(input_shape)
            input_shape = (input_shape[0], layer.units)
        super().build(input_shape)

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        """Run the network on inputs."""
        for layer in self.field_layers:
            inputs = layer(inputs)
        return inputs



from . import items, smap
import random


def test_no_associations():
    """Test number of associations of a map."""
    assert smap.no_associations() == 4, f"Expected 4, got {smap.no_associations()}"


def test_association_sizes():
    """Test number of similarities used by each association."""
    expected = [1, 2, 2, 1]
    assert (
        smap.association_sizes() == expected
    ), f"Expected {expected}, got {smap.association_sizes()}"


def test_association_offsets():
    """Test association offsets."""
    expected = [0, 1, 3, 5]
    assert (
        smap.association_offsets() == expected
    ), f"Expected {expected}, got {smap.association_offsets()}"


def test_similarity_map_length():
    """Test similarity map length."""
    expected = len(items)
    assert len(smap.lcols) == expected, f"Expected {expected}, got {len(smap.lcols)}"


def test_similarity_map_iter():
    """Test similarity map iteration."""
    assert list(smap) == items, f"Expected: \n{items},\n got {list(smap)}"


def test_similarity_map_item():
    """Test similarity map item."""
    i = random.sample(range(len(items)), 1)[0]
    expected = items[i]
    assert smap[i] == expected, f"Expected {expected}, got {smap[i]}"



"""
Local explainability module.

This module provides customized local explainability functionality for entity
matching tasks.
"""

from neer_match.data_generator import DataGenerator
from neer_match.matching_model import (
    _matching_model_or_raise,
    DLMatchingModel,
    NSMatchingModel,
)
import numpy as np
import pandas as pd
import tensorflow as tf
import typing


def _xindices_or_raise(xindices: typing.List[int]) -> None:
    if (
        not isinstance(xindices, list)
        or len(xindices) != 2
        or not all([isinstance(i, int) for i in xindices])
    ):
        raise ValueError("The xindices argument must be a list of two integers")


def lime(
    model: typing.Union[DLMatchingModel, NSMatchingModel],
    left: pd.DataFrame,
    right: pd.DataFrame,
    xindices: typing.List[int],
    n: int = 100,
) -> pd.DataFrame:
    """Calculate local interpretable model-agnostic explanations.

    Generates a sample of n instances around the pair at the given indices
    and fits a weighted linear model to explain the prediction of the model.

    Args:
        model: The matching model to use.
        left: The left DataFrame.
        right: The right DataFrame.
        xindices: The indices of the pair to explain.
        n: The number of samples to generate.
    """
    _matching_model_or_raise(model)
    _xindices_or_raise(xindices)
    if not isinstance(n, int) or n < len(model.similarity_map):
        raise ValueError(
            f"The n argument must be a positive integer >= {len(model.similarity_map)}."
        )
    # The data frame arguments are validated by the DataGenerator
    generator = DataGenerator(
        model.record_pair_network.similarity_map,
        left,
        right,
    )
    xl = left.iloc[xindices[0]]
    xr = right.iloc[xindices[1]]
    x = generator.similarity_encoder.encode_as_matrix(xl, xr)
    sum_x = 0.0
    sum_xx = 0.0
    for features in generator:
        sum_x += tf.concat(
            [tf.reduce_sum(f, axis=0) for k, f in generator[0].items()], 0
        )
        sum_xx += tf.concat(
            [tf.reduce_sum(f**2, axis=0) for k, f in generator[0].items()], 0
        )
    mean = sum_x / generator.no_pairs()
    stddev = tf.sqrt(sum_xx / generator.no_pairs() - mean**2)
    sample = tf.random.normal(
        (n, len(generator.similarity_map)),
        mean=mean,
        stddev=stddev,
        dtype=tf.dtypes.float64,
    )
    sample = tf.clip_by_value(sample, 0.0, 1.0)
    dist = sample - x
    weights = tf.exp(-tf.reduce_sum(dist**2, axis=1) / 2.0)

    begins = generator.similarity_encoder.assoc_begin
    ends = generator.similarity_encoder.assoc_end
    names = generator.similarity_map.association_names()
    features = {k: sample[:, begins[i] : ends[i]] for i, k in enumerate(names)}
    y = tf.cast(model(features), tf.dtypes.float64)

    weights = tf.linalg.diag(weights)
    X = tf.concat([tf.ones((n, 1), dtype=tf.dtypes.float64), sample], axis=1)
    u = tf.tensordot(X, weights, axes=[[0], [1]])
    p1 = tf.tensordot(u, X, axes=[[1], [0]])
    vcov = tf.linalg.inv(p1)
    p2 = tf.tensordot(u, y, axes=[[1], [0]])
    beta = tf.tensordot(vcov, p2, axes=[[1], [0]])

    return pd.DataFrame(
        {
            "feature": ["intercept"] + generator.similarity_map.keys(),
            "coef": beta.numpy().flatten(),
            "stderr": tf.sqrt(tf.linalg.diag_part(vcov)).numpy(),
        }
    )


def shap(
    model: typing.Union[DLMatchingModel, NSMatchingModel],
    left: pd.DataFrame,
    right: pd.DataFrame,
    xindices: typing.List[int],
    xkey: str,
    iterations: int = 100,
) -> float:
    """Calculate the Shapley value of a key.

    Estimates the Shapley value of a key for a pair of instances by sampling
    random sets of features and comparing the model predictions with and without
    the key.

    Args:
        model: The matching model to use.
        left: The left DataFrame.
        right: The right DataFrame.
        xindices: The indices of the pair to explain.
        xkey: The key for which to calculate the Shapley value.
        iterations: The number of iterations to use.
    """
    _matching_model_or_raise(model)
    _xindices_or_raise(xindices)
    generator = DataGenerator(
        model.record_pair_network.similarity_map,
        left,
        right,
    )

    offsets = np.array(model.similarity_map.association_offsets())
    names = model.similarity_map.association_names()
    n_positions = {}
    for i, key in enumerate(generator.similarity_map.keys()):
        if key == xkey:
            continue
        lower_than = offsets[offsets <= i]
        n_positions[i] = (len(lower_than) - 1, i - lower_than[-1])

    xl = left.iloc[xindices[0]]
    xr = right.iloc[xindices[1]]
    x_feature = generator.similarity_encoder(xl, xr)
    x_feature = {k: v for k, v in zip(names, x_feature)}
    x_aindex = generator.similarity_map.keys().index(xkey)
    xlower_than = offsets[offsets <= x_aindex]
    x_name = names[len(xlower_than) - 1]
    x_sindex = x_aindex - xlower_than[-1]

    n_count = len(n_positions)
    n_indices = list(n_positions.keys())
    shapley = 0.0
    for i in range(iterations):
        count = tf.random.uniform((), maxval=n_count, dtype=tf.dtypes.int32).numpy()
        indices = tf.random.shuffle(n_indices)[:count].numpy()
        positions = {pos: val for pos, val in n_positions.items() if pos in indices}
        zbatch = generator[
            tf.random.uniform((), maxval=len(generator), dtype=tf.dtypes.int32).numpy()
        ]
        zbindex = tf.random.uniform(
            (), maxval=generator.batch_size, dtype=tf.dtypes.int32
        ).numpy()
        z_feature = {k: v[zbindex].reshape(1, -1) for k, v in zbatch.items()}
        xp = {k: v.copy() for k, v in x_feature.items()}
        for pos, (apos, spos) in positions.items():
            name = names[apos]
            xp[name][:, spos] = z_feature[name][:, spos]
        xm = {k: v.copy() for k, v in xp.items()}
        xm[x_name][:, x_sindex] = z_feature[x_name][:, x_sindex]
        fp = model(xp)
        fm = model(xm)
        shapley += (fp - fm).numpy()[0, 0]
    shapley /= iterations

    return shapley



from . import dl_model, left_short, right_short
from neer_match.local_explainability import lime, shap
import pandas as pd


def test_lime():
    """Test LIME explainability"""
    results = lime(dl_model, left_short, right_short, [0, 1], n=8)
    assert isinstance(results, pd.DataFrame), "Failed to explain with LIME."
    no_expected_rows = len(dl_model.similarity_map) + 1
    assert (
        results.shape[0] == no_expected_rows
    ), f"Expected {len(left_short)} explanations, got {results.shape[0]}"


def test_shap():
    """Test SHAP explainability."""
    results = shap(
        dl_model,
        left_short,
        right_short,
        [0, 1],
        "title_jaro_winkler",
        iterations=1,
    )
    assert isinstance(results, float), "Failed to explain with SHAP."



from . import ns_model, left_short, right_short
from neer_match.global_explainability import (
    accumulated_local_effect_plot,
    partial_dependence,
    partial_dependence_feature_importance,
    partial_dependence_plot,
)
import numpy as np
import matplotlib


def test_partial_dependence():
    """Test partial dependence."""
    interp_n = 2
    result = partial_dependence(
        ns_model, left_short, right_short, "title_jaro_winkler", n=interp_n
    )
    assert isinstance(result, np.ndarray), (
        "Expected float result. " f"Instead got {type(result)}"
    )
    expected = (interp_n,)
    assert result.shape == expected, (
        f"Expected shape {expected}. " f"Instead got {result.shape}"
    )
    result = partial_dependence_feature_importance(
        ns_model, left_short, right_short, "title_jaro_winkler", n=interp_n
    )
    assert isinstance(result, np.float32), (
        "Expected float result. " f"Instead got {type(result)}"
    )
    assert result >= 0, "Expected positive result. " f"Instead got {result}"
    fig = partial_dependence_plot(
        ns_model, left_short, right_short, "title_jaro_winkler", n=interp_n
    )
    assert isinstance(
        fig, matplotlib.figure.Figure
    ), "Failed to generate partial dependence plot."


def test_accumulated_local_effect():
    """Test accumulated local effect."""
    simil_k = 2
    fig = accumulated_local_effect_plot(
        ns_model, left_short, right_short, "title_jaro_winkler", n=2, k=simil_k
    )
    assert isinstance(
        fig, matplotlib.figure.Figure
    ), "Failed to generate accumulated local effect plot."
